/*
 * Copyright (C) 2019 Weber Yann
 * 
 * This file is part of PyFCGI.
 * 
 * PyFCGI is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PyFCGI is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with PyFCGI.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "responder.h"

void init_context()
{
	PyFCGI_conf.context.pid = getpid();
	PyFCGI_conf.context.ppid = getppid();

	if(pyfcgi_IPC_create(IPC_WSTATE | IPC_WREQS | IPC_SEMST) < 0)
	{
		pyfcgi_log(LOG_ALERT, "Pool handler process is unable to create WSTATE SEM");
		sleep(1);
		clean_exit(PYFCGI_FATAL);
	}
	/*
	if(pyfcgi_IPC_create(IPC_WSTATE) < 0)
	{
		pyfcgi_log(LOG_ALERT, "Pool handler process is unable to create WSTATE SEM");
		sleep(1);
		clean_exit(PYFCGI_FATAL);
	}
	if(pyfcgi_IPC_init(IPC_WREQS | IPC_SEMST) < 0)
	{
		pyfcgi_log(LOG_ALERT, "Pool handler process is unable to init IPC components");
		sleep(1);
		clean_exit(PYFCGI_FATAL);
	}
	*/
}

int responder_loop()
{
	unsigned int n_wrk, wanted_n, n;
	pid_t *wrk_pids;
	int err;
	int status;
	pid_t ret;
	/**@brief poll timeout */
	struct timespec timeout;
	/**@brief watchdog timeout */
	struct timespec pool_timeout;
	time_t idle_start, busy_start;
	short idle, busy;
	struct sigaction act;
	char *statusstr;

	act.sa_handler = pool_sighandler;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGTERM);
	act.sa_flags = 0;
	act.sa_restorer = NULL;

	if(sigaction(SIGINT, &act, NULL))
	{
		perror("Sigaction error for pool process");
		exit(PYFCGI_FATAL);
	}

	timeout.tv_sec = 0;
	timeout.tv_nsec = 100000000;
	idle = busy = 0;

	pyfcgi_logger_set_ident("Workpool");

	if(PyFCGI_conf.pool_timeout)
	{
		pool_timeout.tv_nsec = 0;
		pool_timeout.tv_sec = PyFCGI_conf.pool_timeout;
		pyfcgi_wd_init(pool_wd_sighandler, &pool_timeout);
	}

	pyfcgi_log(LOG_INFO, "Preparing workers");

	init_context();

	pyfcgi_wd_arm();

	PyFCGI_conf.context.wrk_pids = &wrk_pids;
	PyFCGI_conf.context.n_wrk = 0;
	wrk_pids = malloc(sizeof(int) * PyFCGI_conf.max_wrk);
	if(!wrk_pids)
	{
		err = errno;
		pyfcgi_log(	LOG_ALERT,
			"Unable to allocate memory for childs PID : %s",
			strerror(err));
		clean_exit(err);
	}
	bzero(wrk_pids, sizeof(int) * PyFCGI_conf.max_wrk);

	wanted_n = PyFCGI_conf.min_wrk;
	n_wrk = 0;
	// prespawning minimum worker count
	for(n_wrk=0; n_wrk < wanted_n; n_wrk++)
	{
		wrk_pids[n_wrk] = spawn(n_wrk);
		PyFCGI_conf.context.n_wrk = n_wrk;
	}
	//Wait at least for a process to be ready
	while(!pyfcgi_pool_idle(&timeout));

	// main loop, taking care to restart terminated workers, 
	// spawn new one if needed, etc.
	while(1)
	{
		pyfcgi_wd_arm();
		PyFCGI_conf.context.n_wrk = n_wrk;
		if( (ret = waitpid(0, &status, WNOHANG)) )
		{
			if(ret < 0)
			{
				//TODO : error
			}
			for(n=0; n<n_wrk; n++)
			{
				if(wrk_pids[n] == ret)
				{
					break;
				}
			}
			if(n == n_wrk)
			{
				pyfcgi_log(LOG_WARNING,
				       "Child %d stopped but was notregistered",
				       ret);
				continue;
			}
			if(WIFSIGNALED(status))
			{
				if(WTERMSIG(status) == 11)
				{
					pyfcgi_log(LOG_ALERT,
						"Worker[%d] segfault !",
						n);
				}
				else
				{
					pyfcgi_log(LOG_ALERT,
						"Worker[%d] terminated by signal %s(%d)",
						n, strsignal(WTERMSIG(status)),
						WTERMSIG(status));
				}
			}
			if(WEXITSTATUS(status))
			{
				statusstr = status2str(WEXITSTATUS(status));
				pyfcgi_log((WEXITSTATUS(status)&PYFCGI_FATAL)?
						LOG_ALERT:LOG_WARNING,
					"Worker[%d] exited with status %s",
					n, statusstr);
				free(statusstr);
			}
			if(!status)
			{
				pyfcgi_log(LOG_INFO,
				       "Worker[%d] PID %d exited normally",
				       n, wrk_pids[n]);
			}

			// respawn on same slot
			pyfcgi_log(LOG_DEBUG, "respawning worker #%d", n);
			wrk_pids[n] = spawn(n);

			continue;
		}

		// Check if the pool is idle or busy
		if(pyfcgi_pool_idle(&timeout))
		{
			// workers idle
			busy = 0;
			if(!idle)
			{
				idle = 1;
				idle_start = time(NULL);	
			}
			else if((time(NULL) - idle_start) > PyFCGI_conf.worker_gc_timeout &&
				wanted_n > PyFCGI_conf.min_wrk
				&& n_wrk - wanted_n < 2)
			{
				wanted_n--;
				idle = 0;
			}
		}
		else
		{
			idle = 0;
			if(!busy)
			{
				busy = 1;
				busy_start = time(NULL);
			}
			else if(time(NULL) - busy_start > 0 &&
				wanted_n < PyFCGI_conf.max_wrk)
			{
				pyfcgi_log( LOG_DEBUG,
					"All workers busy, spawning a new one");
				n = n_wrk;
				n_wrk++;
				wanted_n = n_wrk;
				wrk_pids[n] = spawn(n);
				if(!PyFCGI_conf.worker_fast_spawn)
				{
					busy_start = time(NULL);
				}
			}
		}

		// Stopping & deleting useless childs
		if(wanted_n < n_wrk && idle)
		{	// need to shift the list and dec n_wrk
			busy = 0;
			n_wrk--;
			kill(wrk_pids[n_wrk], SIGTERM);
			nanosleep(&timeout, NULL);
			if( (ret = waitpid(wrk_pids[n_wrk], &status, WNOHANG)) < 0 )
			{
				pyfcgi_log(LOG_ERR, "Pool idle since %ds but unable to kill child %d (PID %d)",
					PyFCGI_conf.worker_gc_timeout,
					n_wrk, wrk_pids[n_wrk]);
				kill(wrk_pids[n_wrk], SIGKILL);
			}
			else
			{
				pyfcgi_log(LOG_INFO, "Pool idle since %ds : worker[%d](%d) killed",
					PyFCGI_conf.worker_gc_timeout,
					n_wrk, wrk_pids[n_wrk]);
			}
			idle = 0;
			continue;
		}


		nanosleep(&timeout, NULL);
	}
	
	pyfcgi_wd_arm();
	//Debug wait & exit
	for(; n_wrk != 0; n_wrk--)
	{
		waitpid(wrk_pids[n_wrk], &status, 0);
		pyfcgi_log(LOG_DEBUG, "Child %d stopped with status %d",
		       wrk_pids[n_wrk], status);
		PyFCGI_conf.context.n_wrk = n_wrk;
	}
		//printf("Content-Type: text/html\r\n\r\nHello world !\n");
	pyfcgi_wd_stop();
	pyfcgi_log(LOG_INFO,"Child workers stoped, stopping responder");
	exit(0);
}

pid_t spawn(int wrk_id)
{
	pid_t res;
	struct timespec wd_timeout;
	struct sigaction act;
	char ident[128];

	act.sa_handler = worker_sighandler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	act.sa_restorer = NULL;


	res = fork();
	if(res == -1)
	{
		pyfcgi_log(LOG_ERR, "Fork fails for worker #%d : %s",
			wrk_id, strerror(errno));
		return -1;
	}
	else if(!res)
	{
		// Child process
		PyFCGI_conf.context.ppid = PyFCGI_conf.context.pid;
		PyFCGI_conf.context.pid = getpid();
		snprintf(ident, 128, "Worker%2d", wrk_id);
		pyfcgi_logger_set_ident(ident);
		// Init IPC components
		if(pyfcgi_IPC_init(IPC_WSTATE | IPC_WREQS) < 0)
		{
			pyfcgi_log(LOG_ALERT, "Unable to initialize semaphore when spawning process...");
			exit(PYFCGI_FATAL);
		}
		// Set handler for SIGINT & SIGTERM
		if(sigaction(SIGINT, &act, NULL))
		{
			perror("Sigaction error for pool process");
			exit(PYFCGI_FATAL);
		}
		if(sigaction(SIGTERM, &act, NULL))
		{
			perror("Sigaction2 error for pool process");
			exit(PYFCGI_FATAL);
		}
		// Set watchdog
		if(PyFCGI_conf.worker_timeout)
		{
			wd_timeout.tv_nsec = 0;
			wd_timeout.tv_sec = PyFCGI_conf.worker_timeout;
			pyfcgi_wd_init(worker_sigalrmhandler, &wd_timeout);
		}

		if(PyFCGI_conf.pep333)
		{
			exit(work333(wrk_id));
		}
		else
		{
			exit(work(wrk_id));
		}
	}
	pyfcgi_IPC_init(IPC_WSTATE | IPC_WREQS | IPC_SEMST);
	// Sleep to avoid spawning like hell thinking all workers are
	// busy. Let some time to this one to go up...
	// TODO: find a better way to avoid spawning to max_wrk
	//nanosleep(&timeout, NULL);
	pyfcgi_log(	LOG_INFO,
		"Worker #%d spawned with PID %d", wrk_id, res);
	return res;
}

int pyfcgi_pool_state()
{
	int err, res;
	if(sem_getvalue(PyFCGI_SEM(SEM_WSTATE).sem, &res) < 0)
	{
		err = errno;
		pyfcgi_log(LOG_ALERT, "Unable to read WSTATE semaphore value : %s",
			strerror(err));
		clean_exit(PYFCGI_FATAL);
	}
	return res;
}

int pyfcgi_pool_idle(const struct timespec *timeout)
{
	int err;
	struct timespec abs_timeout;

	if(clock_gettime(CLOCK_REALTIME_COARSE, &abs_timeout) < 0)
	{
		//clock error
		pyfcgi_log(LOG_WARNING, "Unable to fetch asbtime for WSTATE sem_timedwait : %s",
			strerror(errno));
	}
	abs_timeout.tv_sec += timeout->tv_sec;
	if(abs_timeout.tv_nsec + timeout->tv_nsec > 999999999)
	{
		abs_timeout.tv_nsec = abs_timeout.tv_nsec + timeout->tv_nsec - 999999999;
		abs_timeout.tv_sec +=1;
	}
	else
	{
		abs_timeout.tv_nsec = timeout->tv_nsec;
	}

	if(sem_timedwait(PyFCGI_SEM(SEM_WSTATE).sem, &abs_timeout) < 0)
	{
		err = errno;
		switch(err)
		{
			case ETIMEDOUT:
			case EAGAIN:
				return 0; //busy
			case EINVAL:
				sleep(1);
				return 1;
			default:
				pyfcgi_log(LOG_ALERT, "Unable to wait WSTATE sem : %s",
					strerror(err));
				clean_exit(PYFCGI_FATAL);
		}
	}
	sem_post(PyFCGI_SEM(SEM_WSTATE).sem); //Hope no worker fails to set busy...
	return 1; //idle
}

void clean_exit(int status)
{
	pyfcgi_IPC_close(IPC_WSTATE | IPC_WREQS | IPC_SEMST | IPC_SHMST);
	//temporarly destroy everything....
	pyfcgi_IPC_destroy(IPC_WSTATE | IPC_WREQS | IPC_SEMST | IPC_SHMST);
	//pyfcgi_IPC_destroy(IPC_WSTATE);
	exit(status);
}

void pool_sighandler(int signum)
{
	unsigned int i;
	struct timespec req;
	req.tv_sec = 0;
	req.tv_nsec = 200000000; //0.2s
	if(PyFCGI_conf.context.n_wrk < 1) { clean_exit(0); }

	for(i=0; i<PyFCGI_conf.context.n_wrk; i++)
	{
		pyfcgi_log(LOG_INFO, "Sending SIGTERM to child #%d (pid %d)",
			i,(*PyFCGI_conf.context.wrk_pids)[i]);
		kill((*PyFCGI_conf.context.wrk_pids)[i], SIGTERM);
		nanosleep(&req, NULL); //waiting 0.2s
	}
	for(i=0; i<PyFCGI_conf.context.n_wrk; i++)
	{
		if(kill((*PyFCGI_conf.context.wrk_pids)[i], SIGCONT))
		{
			pyfcgi_log(LOG_INFO, "Sending SIGKILL to child %d", i);
			kill((*PyFCGI_conf.context.wrk_pids)[i], SIGKILL);
		}
		
	}
	clean_exit(0);
}

void pool_wd_sighandler(int signum)
{
	unsigned int i;
	pyfcgi_log(LOG_ALERT, "Worker pool timeout ! Attempt to kill all childs");
	for(i=0; i<PyFCGI_conf.context.n_wrk; i++)
	{
		pyfcgi_log(LOG_ALERT, "Child[%d] PID %d", i, (*PyFCGI_conf.context.wrk_pids)[i]);
		kill((*PyFCGI_conf.context.wrk_pids)[i], SIGALRM);
	}
	while(PyFCGI_conf.context.n_wrk)
	{
		kill((*PyFCGI_conf.context.wrk_pids)[PyFCGI_conf.context.n_wrk], SIGALRM);
		PyFCGI_conf.context.n_wrk--;
	}
	pyfcgi_wd_stop();
	kill(PyFCGI_conf.context.pid, SIGTERM);
	clean_exit(PYFCGI_TIMEOUT);
	exit(PYFCGI_TIMEOUT);
}
